<?php

include_once __DIR__.'/Filename.php';
include_once __DIR__.'/FileNotFoundException.php';
include_once __DIR__.'/FolderNotFoundException.php';
include_once __DIR__.'/ItemAlreadyExistsException.php';
include_once __DIR__.'/ModifiedDateException.php';
include_once __DIR__.'/NameException.php';
include_once __DIR__.'/ZviniAuthenticationException.php';
include_once __DIR__.'/ZviniConnectionException.php';

class ZviniProxy {

    public $canSearch = false;
    public $type = 'zvini-folder';
    public $isProxy = true;

    public $name;
    public $address;
    public $username;
    public $password;

    private $cookies = '';

    function __construct ($name, $address, $username, $password) {
        if (!Filename::isValid($name) || $address === '') {
            throw new NameException;
        }
        if ($username === '' || $password === '') {
            throw new ZviniAuthenticationException;
        }
        if (substr($address, -1) !== '/') $address .= '/';
        $this->name = $name;
        $this->address = $address;
        $this->username = $username;
        $this->password = $password;
        $this->authenticate();
    }

    private function authenticate () {
        $response = $this->call('session/authenticate', [
            'username' => $this->username,
            'password' => $this->password,
        ]);
        if ($response === 'INVALID_USERNAME' || $response === 'INVALID_PASSWORD' ||
            $response === 'INVALID_USERNAME_OR_PASSWORD') {

            throw new ZviniAuthenticationException;

        }
    }

    function call ($method, $params) {
        $raw_response = $this->download($method, $params);
        $response = json_decode($raw_response);
        if ($response === null) throw new ZviniConnectionException;
        return $response;
    }

    private function callAuthenticated ($method, $params) {
        $params['session_auth'] = '1';
        $response = $this->call($method, $params);
        if ($response !== 'NOT_SIGNED_IN') return $response;
        $this->authenticate();
        return $this->call($method, $params);
    }

    private function chdir ($folder_id, $splitPaths) {
        $passedPaths = [];
        foreach ($splitPaths as $splitPath) {
            $passedPaths[] = $splitPath;
            $folders = $this->listFolders($folder_id);
            $found = false;
            foreach ($folders as $folder) {
                if ($folder->name !== $splitPath) continue;
                $folder_id = $folder->id;
                $found = true;
                break;
            }
            if (!$found) {
                throw new FolderNotFoundException(Path::join($passedPaths));
            }
        }
        return $folder_id;
    }

    function createFolder ($path) {
        $splitPaths = Path::split($path);
        $folderName = array_pop($splitPaths);
        $folder_id = $this->chdir(0, $splitPaths);
        $response = $this->callAuthenticated('folder/add', [
            'parent_id' => $folder_id,
            'name' => $folderName,
        ]);
        if ($response === 'FOLDER_ALREADY_EXISTS') {
            throw new ItemAlreadyExistsException($path);
        }
    }

    private function download ($method, $params) {

        $cookie_file = tempnam(sys_get_temp_dir(), 'zvini_cookies');
        file_put_contents($cookie_file, $this->cookies);

        $ch = curl_init();
        curl_setopt_array($ch, [
            CURLOPT_URL => "{$this->address}/api-call/$method",
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_POSTFIELDS => $params,
            CURLOPT_COOKIEFILE => $cookie_file,
            CURLOPT_COOKIEJAR => $cookie_file,
        ]);
        $response = curl_exec($ch);
        if ($response === false) throw new ZviniConnectionException;
        curl_close($ch);

        $this->cookies = file_get_contents($cookie_file);
        unlink($cookie_file);
        return $response;

    }

    function getFileContent ($path) {
        $splitPaths = Path::split($path);
        $fileName = array_pop($splitPaths);
        $folder_id = $this->chdir(0, $splitPaths);
        foreach ($this->listFiles($folder_id) as $file) {
            if ($file->name !== $fileName) continue;
            return [
                'content' => $this->download('file/download', [
                    'session_auth' => '1',
                    'id' => $file->id,
                ]),
                'mtime' => $file->insert_time,
            ];
        }
        throw new FileNotFoundException($path);
    }

    function getItem ($path) {

        if ($path === '') return $this->listFoldersAndFiles(0);
        $splitPaths = Path::split($path);
        $name = array_pop($splitPaths);
        $folder_id = $this->chdir(0, $splitPaths);

        foreach ($this->listFolders($folder_id) as $folder) {
            if ($folder->name !== $name) continue;
            return $this->listFoldersAndFiles($folder->id);
        }

        foreach ($this->listFiles($folder_id) as $file) {
            if ($file->name !== $name) continue;
            $content = $this->download('file/download', [
                'session_auth' => '1',
                'id' => $file->id,
            ]);
            return new File($name, $content);
        }

        throw new FileNotFoundException($path);

    }

    function index ($path) {
        $splitPaths = Path::split($path);
        $folder_id = $this->chdir(0, $splitPaths);
        return $this->listFoldersAndFiles($folder_id);
    }

    private function listFiles ($folder_id) {
        return $this->callAuthenticated('file/list', [
            'parent_id' => $folder_id,
        ]);
    }

    private function listFolders ($folder_id) {
        return $this->callAuthenticated('folder/list', [
            'parent_id' => $folder_id,
        ]);
    }

    private function listFoldersAndFiles ($folder_id) {
        $items = [];
        foreach ($this->listFolders($folder_id) as $folder) {
            $items[] = [
                'type' => 'folder',
                'name' => $folder->name,
            ];
        }
        foreach ($this->listFiles($folder_id) as $file) {
            $items[] = [
                'type' => 'file',
                'name' => $file->name,
            ];
        }
        return $items;
    }

    function putFileContent ($path, $content, $overwrite, &$mtime = null) {
        $splitPaths = Path::split($path);
        $fileName = array_pop($splitPaths);
        $folder_id = $this->chdir(0, $splitPaths);
        try {

            $tempnam = sys_tempnam('ftpupload');
            file_put_contents($tempnam, $content);

            $found_file = null;
            foreach ($this->listFiles($folder_id) as $file) {
                if ($file->name !== $fileName) continue;
                $found_file = $file;
                break;
            }

            if ($found_file !== null) {

                if (!$overwrite) throw new ItemAlreadyExistsException($path);

                if ($mtime) {
                    $oldMtime = $file->insert_time;
                    if ($oldMtime > $mtime) {
                        throw new ModifiedDateException;
                    }
                }

                $this->callAuthenticated('file/delete', ['id' => $file->id]);

            }

            $id = $this->callAuthenticated('file/add', [
                'parent_id' => $folder_id,
                'name' => $fileName,
                'file' => new CURLFile($tempnam),
            ]);
            if ($id === 'FILE_ALREADY_EXISTS') {
                throw new ItemAlreadyExistsException($path);
            }

            $file = $this->callAuthenticated('file/get', ['id' => $id]);

            $mtime = $file->insert_time;
            unlink($tempnam);

        } catch (Exception $e) {
            unlink($tempnam);
            throw $e;
        }
    }

    function rename ($path, $newName) {

        $splitPaths = Path::split($path);
        $name = array_pop($splitPaths);
        $folder_id = $this->chdir(0, $splitPaths);

        foreach ($this->listFolders($folder_id) as $folder) {
            if ($folder->name !== $name) continue;
            return $this->callAuthenticated('folder/rename', [
                'id' => $folder->id,
                'name' => $newName,
            ]);
        }

        foreach ($this->listFiles($folder_id) as $file) {
            if ($file->name !== $name) continue;
            $content = $this->callAuthenticated('file/rename', [
                'id' => $file->id,
                'name' => $newName,
            ]);
            return new File($name, $content);
        }

        throw new FileNotFoundException($path);

    }

    function remove ($path) {

        $splitPaths = Path::split($path);
        $name = array_pop($splitPaths);
        $folder_id = $this->chdir(0, $splitPaths);

        foreach ($this->listFolders($folder_id) as $folder) {
            if ($folder->name !== $name) continue;
            return $this->callAuthenticated('folder/delete', [
                'id' => $folder->id,
            ]);
        }

        foreach ($this->listFiles($folder_id) as $file) {
            if ($file->name !== $name) continue;
            $content = $this->callAuthenticated('file/delete', [
                'id' => $file->id,
            ]);
            return new File($name, $content);
        }

        throw new FileNotFoundException($path);

    }

    private function sessionResume () {
        if ($this->session_id) {
            session_name($this->session_name);
            session_id($this->session_id);
            session_start();
            $this->session_id = null;
            $this->session_name = null;
        }
    }

    private function sessionSuspend () {
        $session_id = session_id();
        if ($session_id) {
            $this->session_id = $session_id;
            $this->session_name = session_name();
            session_commit();
        }
    }

    function toClientJson () {
        return [
            'type' => 'zvini-folder',
            'name' => $this->name,
        ];
    }

}
